
  'AD9850 VFO Controller V9
  '(c)Peter Carnegie - GM8JCF - 2014
'=================
'Rotary Encoder - exactly as per page 22 Micromite User Manual
'RA -> 2
'RB -> 3
'===================
'Ebay AD9850 Module see http://www.alhin.de/arduino/index.php?n=7
'3 (FQUD) -> 18
'4 (Data) -> 17
'2 (WCLK) -> 16
'5        -> GND
'6        -> GND
'1        -> Vcc (3V3)
'20       -> Vcc (3V3)
'19       -> 4K7 -> Vcc(3V3)
'18       -> 4K7 -> Vcc(3V3)
'17       -> Gnd
'11       -> Gnd
'==================
'LCD - 16x2 lines - see page 19 Micromite USer Manual
'4  (RS) -> 22
'6  (EN) -> 21
'14 (D7) -> 26
'13 (D6) -> 25
'12 (D5) -> 24
'11 (D4) -> 23
'==================
'Future rekon -with I2C display
'1  MCLR
'2  ADC-1
'3  ADC-2
'4  ADC-3
'5  ADC-4 Battery Voltage
'6  RA - Rotary Encoder
'7  RB - Rotary Encoder
'8  Ground
'9  FQUD - AD9850
'10 WCLK - AD9850
'11 Console Tx
'12 Console Rx
'13 Power
'14 SData - AD9850
'15 KPD-R1
'16 Tune Step Change/Wake Up
'17 I2C-SCL - LCD
'18 I2C-SDA - LCD
'19 Gnd
'20 Vcap
'21 KPD-R2
'22 KPD-R3
'23 KPD-R4
'24 KPD-C1
'25 KPD-C2
'26 KPD-C3
'27 Analog Gnd
'28 Analog Pwr
'================

  'eBay AD9850 Modules use 125,000,000 Hz Clocks
  'AD9850 Fout= (? Phase  CLKIN)/2^32 where Fout=32 bit tuning word
  'TuningWord= Fout * 2^32 / CLKIN = Fout * (2^32/125,000,000)
  '2^32/125,000,000 = 34.359738368 - VB6 Double Precision
  'Using Integer only arithmetic, multiply by 1,000,000
  'DivConst= 34359738
  'We then divide the result of TuningWord=DivConst * Fout by 1,000,000

  'Initialise the BCD Library
  InitBCDLib

  Dim DivConst(7)
  'Load it up in BCD
  LoadBCD DivConst(0),"34359738"

  'Cur Freq in BCD
  Dim CurFreq(7)
  'Load it up in BCD
  LoadBCD CurFreq(0),"10000000"

  'Define DDS Control pins
  FQUD=18
  SData=17
  WCLK=16

  'Initialise the LCD Display
  DispLen=16
  '26=D7, 25=D6, 24=d5, 23=d4
  '22=RS, 21=EN
  LCD init 23,24,25,26,22,21

  'Rotary Encoder Input pins - see page 22 uMite User Manual
  Dim RA,RB
  RB=2  'If you use mechanical encoder
  RA=3  'If you use mechanical encoder
  RSW=4 'Roll switch
  SetPin RB, Din
  SetPin RA, INTH,  RInt
  SETPIN RSW, INTL, Roll
  

  'Variable incremented/decremented by the Interrupt handling
  'routine for the Rotary Encoder
  Dim Value,LastValue

  'Raw Tuning Word, ie scaled by 1E6
  Dim RawTWord(7)

  'Tuning Word scaled back to real world units, ie divide by 1E6
  Dim TWord(7)

  'Index of which Step size is in force
  Dim CurStepIndex

  'Startup with 100Hz tuning steps
  CurStepIndex=2

  'Current Step in TWord units
  Dim CurStepTW(7)
  
  'And we to show the step  
  DIM DispStep$(7) Length 6
  DispStep$(0)="   1Hz"
  DispStep$(1)="  10Hz"
  DispStep$(2)=" 100Hz"
  DispStep$(3)="  1kHz"
  DispStep$(4)=" 10kHz"
  DispStep$(5)="100kHz"
  DispStep$(6)="  1MHz"
  
  'Array of Possible Step Sizes in Hz
  Dim StepHz(6)
  StepHz(0)=1
  StepHz(1)=10
  StepHz(2)=100
  StepHz(3)=1000
  StepHz(4)=10000 
  StepHZ(5)=100000
  'What to Add/Subtract from the Scaled tuning Word
  'for each Encoder Up/Down
  Dim StepTWord$(6) Length 24
  'These TuningWord sizes were calculated using Double precision arithmetic
  'in VB6
  StepTWord$(0)="34359738"     '1Hz Step
  StepTWord$(1)="343597384"    '10Hz
  StepTWord$(2)="3435973837"   '100Hz 3435973836.8
  StepTWord$(3)="34359738368"  '1000Hz
  StepTWord$(4)="343597383680"   '10000Hz 343597383680 
  StepTWord$(5)="3435973836800"  '100000Hz 3435973836800 

  'Load the CurStepTW with the startup Step in TW
  LoadBCD CurStepTW(0),StepTWord$(CurStepIndex)

  'Calculate Starting TuningWord
  CopyBCD DivConst(0),BCDA(0)
  CopyBCD CurFreq(0),BCDB(0)
  'Calculate the Tuning Word, this one is * 1E6
  MulAB2C

  'Save the RawTWord
  CopyBCD BCDC(0), RawTWord(0)

  'Start with clean BCD Registers
  ClrBCD(BCDA(0))
  ClrBCD(BCDB(0))
  ClrBCD(BCDC(0),15)

  'Go do Main Loop
  Main

End

Sub Main

  'General purpose String used in Main Loop
  Local D$(1) length 24
  Local I,L,Cy

  'Flag when rotary encoder has moved.
  Local Change

  D$(1)="GM8JCF  "+Time$
  SetTick 1002,DOIT
  Memory
  LCD 1, C16, D$(1)

  DispFreq

  'Initialise the I/O & DDS chip
  InitDDS

  'Save Power
  'CPU 5

  'Do for ever !
  Do

    Change=Value-LastValue
    Value=LastValue

    'Any movement ?
    If Change<>0 Then

      'We need all the speed we can get for the calculations
      CPU 48

      'If encoder is turning ANTI-clockwise, Decrement
      'If encoder is turning CLOCKwise, Increment
      If Change<0 Then
        DecCurFreq(StepHz(CurStepIndex))
      Else
        IncCurFreq(StepHz(CurStepIndex))
      EndIf

      'Calculate the change to the Tuning Word
      'If ANTI-clockwise then decrement TuningWORD
      'If CLOCKwise then incrementg TuningWORD
      If Change<0 Then
        'Decrement
        Cy=0
        For I=0 To 7
          If RawTword(I)>=CurStepTW(I)+Cy Then
            RawTword(I)=RawTword(I)-CurStepTW(I) - Cy
            Cy=0
          Else
            RawTword(I)=RawTword(I)+1000 - CurStepTW(I) - Cy
            Cy=1
          EndIf
        Next I
      Else
        'Increment
        Cy=0
        For I=0 To 7
          RawTword(I)=RawTword(I) + CurStepTW(I) + Cy
          If RawTword(I)>999 Then
            Cy=RawTword(I)\1000
            RawTword(I)=RawTword(I) Mod 1000
          Else
            Cy=0
          EndIf
        Next I
      EndIf

      'Save the New Tuning Word for next time
      For I=0 To 7
        BCDC(I)=RawTword(I)
      Next I

      'Round to the nearest 1,000,000
      'ie Add 500,000, then chop off the last 6 digits
      Cy=500
      For I=1 To 3
        BCDC(I)=BCDC(I)+Cy
        If BCDC(I)<1000 Then
          Exit For
        Else
          Cy=BCDC(I)\1000
          BCDC(I)=BCDC(I) Mod 1000
        EndIf
      Next I

      'And make the Actual DDS Tuning Word, ie scale down by 1E6
      'For BCD that just means removing the last 6 digits
      For I=2 To 7
        BCDA(I-2)=BCDC(I)
      Next I
      BCDA(6)=0:BCDA(7)=0

      'Convert BCD to UINT32
      BCDA2Bin

      SendDDSData

      DispFreq

      'Reset change flag for next iteration
      Change=0

      'Slowdown the CPU to save power
      'CPU 5

    EndIf

    'Go check again
  Loop

End Sub

End



'Interrupt Routine
'Here when the Rotary encoder moves
RInt:

  If Pin(RB)=1 Then
    'Clockwise rotation
    Value=Value-1
  Else
    'Anti-Clockwise rotation
    Value=Value+1
  EndIf

  IReturn

'===============
DOIT:
  LCD 1,9,Time$
IReturn
'=========================================================
'Increment CurFreq() by Inc
'
''========================================================= 
  'Increment CurFreq() by Inc 
  ' 
  ' 
SUB IncCurFreq(Inc) 
  LOCAL I,Cy 
   
  Cy=Inc 
  FOR I=0 TO 3  
    CurFreq(I)=CurFreq(I) + Cy 
     
    IF CurFreq(I)<1000 THEN 
      EXIT FOR 
    ELSE 
      Cy=CurFreq(I)\1000 
      CurFreq(I)=CurFreq(I) MOD 1000 
    ENDIF 
     
  NEXT I 
   
END SUB 
   
  '========================================================= 
  'Decrement CurFreq() by Dec 
  ' 
  ' 

SUB DecCurFreq(Dec) 
   
  LOCAL I, Bw 

  Bw=Dec 
  FOR I=0 TO 3  
    CurFreq(I)=CurFreq(I)+1000 - BW 
    BW=1 - (CurFreq(I) \ 1000) 
    CurFreq(I)=CurFreq(I) MOD 1000 
  NEXT I 
   
END SUB 


'=========================================================
'Returns a string representing the BCD Digits from CurFreq()
'eg if CurFreq(0)=123, CurFreq(1)=456,CurFreq(2)=789, then
'CurFreq2ASC$ will return "789456123"
'
Function CurFreq2Asc$()

  Local I,J,A$(1) Length BCDSize*3,B$(1) Length BCDSize*3

  'Frequency is 8 digits long , eg 30,000,000
  For I=2 To 0 Step -1

    'Find First non Zero Triplet
    If CurFreq(I)<>0 Then

      For J=I To 0 Step -1
        B$(1)=B$(1)+Str$(CurFreq(J),3,0,"0")
      Next J

      'Now Remove any leading Zeroes
      For J=1 To 3
        If Peek(VAR B$(1),J)<>&H30 Then
          B$(1)=Mid$(B$(1),J)
          Exit For
        EndIf

      Next J

      Exit For

    EndIf

  Next I

  CurFreq2Asc$=B$(1)

End Function

'=========================================================
'
'
'
Sub InitDDS

  SetPin FQUD, DOUT : Pin(FQUD)=0
  SetPin SData, DOUT : Pin(SData)=0
  SetPin WCLK, Dout : Pin(WCLK)=0
  SetFreq

End Sub

'=========================================================
'
'Fundamental equation for DDS
'Tuningword = FrequencyDesired * (AccumulatorBits / ClockIn)
'For an AD9850, TuningWord = F * ( (2^32) / ClkIn )
'
Sub SetFreq
  Local I,J,TWord(7)

  For I=0 To 7
    BCDA(I)=CurFreq(I)
  Next I

  For I=0 To 7
    BCDB(I)=DivConst(I)
  Next I

  'Calc the Tuning Word
  MulAB2C

  'Save this Long Tuning Word
  For I=0 To 7
    RawTWord(I)=BCDC(I)
  Next I

  'Div by 1E6
  J=0
  ClrBCD(BCDA(0))
  For I=2 To 7
    BCDA(J)=BCDC(I)
    J=J+1
  Next I

  'Convert BCDA() to 8 hex digits, ie 32 bits
  BCDA2Bin
  'Tuning Word is now in BCDC()

  'Send it to out
  SendDDSData

End Sub

'=========================================================
'
'Tuning Word to be sent to the DDS must be in BCDC()
'
'
Sub SendDDSData

  Local I

  BCDC(4)=0

  Pin(FQUD)=0   'Make sure FQUD is Low
  Pin(WCLK)=0   'Clock pin Low
  Pin(Sdata)=0  'Data Pin Low

  'Send the Tuning Word and Control bytes to the AD9850
  For I=0 To 4
    Send8Bits(BCDC(I))
  Next I

  Pulse FQUD,0.05    'Latch the data into the DDS chip

End Sub


'=========================================================
'
'
'
Sub Send8Bits(Byte)
  Local B
  'Make a copy so that we don't damage the callers data
  B=Byte
  For I=0 To 7
    If (B And 1)<>0 Then
      Pin(SData)=1
    Else
      Pin(Sdata)=0
    EndIf
    B=B \ 2
    Pulse WCLK,0.05  'Pulse the clock line
  Next I
End Sub

'=========================================================
'
'
'
Sub DispFreq
  Local CurFreq$(1) Length 8

  CurFreq$(1)=CurFreq2Asc$()

  'Display Frequency on the console
  Print CurFreq$(1)

  'Display Frequency on the LCD on Line 2
  LCD Clear
  LCD 2, 1, CurFreq$(1)
  LCD 2,  11,DispStep$(CurStepIndex)
End Sub

'=========================================================
'
'
'
Sub InitBCDLib(NumDigits)

  If NumDigits=0 Then NumDigits=24

  Dim BCDSize

  'Number of elements per BCD register
  'Each Element holds 3 digits
  BCDSize=(NumDigits\3)
  'The next version of the BCDLib will use BCDSize
  'instead of hard-coded values of 7, 15, 24 etc
  'to allow arbitrary precision arithmetic

  Dim BCDA(BCDSize-1),BCDB(BCDSize-1),BCDC(BCDSize*3)

End Sub


'=========================================================
'BCDC()=BCDA() x BCDB()
'This algorithm works on Triplets of digits
'For optimum speed, make sure the shortest multiplier is in BCDB()
'
Sub MulAB2C

  Local Cy, I, J, K, L, M, N

  'Array for the intermediate multiplication
  Local C1(16)

  For I=0 To 15
    BCDC(I)=0
  Next I

  'Set default length
  M=7
  'Find length of BCDA()
  For I = 7 To 0 Step -1
    If BCDA(I) <> 0 Then
      M=I
      Exit For
    EndIf
  Next I

  'Set default length
  N=7
  'Find Length of BCDB()
  For I = 7 To 0 Step -1
    If BCDB(I) <> 0 Then
      N=I
      Exit For
    EndIf
  Next I

  'Now Do the Long Multiplication of the two numbers
  L = 0
  K = 0

  For I = 0 To M+1
    Cy = 0
    For J = 0 To N+1

      C1(J)=(BCDA(I)*BCDB(J))+Cy

      If C1(J) > 999 Then
        Cy = C1(J) \ 1000
        C1(J) = C1(J) Mod 1000
      Else
        Cy = 0
      EndIf

    Next J

    'Add the partial result to the overall total
    'Add each element of C() to D() result into D()
    Cy = 0

    For J = 0 To 7  '23 \ 3
      K = J + L
      BCDC(K)=BCDC(K)+C1(J)+Cy
      If BCDC(K) > 999 Then
        Cy=BCDC(K)\1000
        BCDC(K)=BCDC(K) Mod 1000
      Else
        Cy = 0
      EndIf
    Next J

    L=L+1

  Next I

End Sub

'=========================================================
'Clear BCD Register
'Dest is Element 0 of a BCD register, eg FREQ(0)
'NumDigits is Number of digits in the BCD register, is Optional
'
Sub ClrBCD(Dest,NumDigits)
  Local I
  If NumDigits=0 Then NumDigits=7

  For I=0 To (NumDigits*4)-1
    Poke VAR Dest,I,0
  Next I

End Sub

'=========================================================
'
'Copy a BCD string into a BCD Register/Destination
'Dest is the first element of a BCD register, eg BCDA(0), CurFreq(0)
'A$ is a string of BCD digits, eg "1234567890"
'
Sub LoadBCD(Dest,A$,NumDigits)
  Local I,J,M
  Local B(15),A2$(28)

  If NumDigits=0 Then NumDigits=7

  'Pad to make sure string is a multiple of 3 digits
  A2$(1)=LPad$(A$,3)

  J = 0
  M=Len(A2$(1))
  For I = M - 2 To 1 Step -3
    B(J) = Val(Mid$(A2$(1), I, 3))
    J = J + 1
  Next I

  For I=0 To (NumDigits*4)-1
    Poke VAR Dest,I,Peek(VAR B(0),I)
  Next I


End Sub

'=========================================================
'Prepend leading 0 to make sure
'string is a multiple of ML digits
'
Function LPad$(A$,ML)
  Local I,J,M
  M=Len(A$)

  If M Mod ML <> 0 Then
    For I=ML To 24 Step ML
      If M<I Then
        '        Print "LPad$:A$";A$ 'DEBUG
        LPad$=String$(I-M,"0")+A$
        Exit For
      EndIf
    Next I
  Else
    LPad$=A$
  EndIf
End Function

'==============================================
'
'Convert BCDA() into an unsigned 32 bit integer
'in BCDC()
'Algorithm process triplets of decimal digits
'
Sub BCDA2Bin

  Local I,T,L,Cy

  For I=0 To 15
    BCDC(I)=0
  Next I

  T=7

  For I=7 To 0 Step -1
    If BCDA(I)<>0 Then
      T=I
      Exit For
    EndIf
  Next I

  'Each triplet of digits are worth 0-999, ie 1000's
  L=1000 ^ (T)
  For I=T To 0 Step -1  '0 To 11

    'Add Next Decimal Digit into the Binary Long word
    BCDC(0)=BCDC(0)+(BCDA(I)*L)

    'Has this caused an overflow
    If BCDC(0)>&HFF Then
      'Yes
      'Now ripple the overflow across the other bytes
      Cy=BCDC(0) And &H7FFFFF00
      BCDC(0)=BCDC(0)And &HFF
      BCDC(1)=BCDC(1)+(Cy\&H100)

      'Has this caused an overflow
      If (BCDC(1)>&HFF) Then
        'Yes
        'Now ripple the overflow across the other bytes
        Cy=BCDC(1) And &H7FFFFF00
        BCDC(1)=BCDC(1) And &HFF
        BCDC(2)=BCDC(2)+(Cy\&H100)

        'Has this caused an overflow
        If (BCDC(2)>&HFF) Then
          'Yes
          'Now ripple the overflow across the other bytes
          Cy=BCDC(2) And &H7FFFFF00
          BCDC(2)=BCDC(2) And &HFF
          BCDC(3)=BCDC(3)+(Cy\&H100)

        EndIf
      EndIf
    EndIf

    L=L\1000

  Next I

  'Was there an overflow
  If (BCDC(3)>&HFF) Then
    Error "OverFlow"
  EndIf

End Sub

'==============================================
'
' General purpose way to move BCD data between source and dest
' arrays. NOT as fast as doing a direct copy using a FOR loop.
'
Sub CopyBCD(Src,Dest,NumDigits)

  Local I
  If NumDigits=0 Then NumDigits=7
  For I=0 To (NumDigits * 4)-1
    Poke VAR Dest,I,Peek(VAR Src,I)
  Next I

End Sub
'=============================================
Roll:
  StepChange
  IRETURN
 '=========================================== 
'Here on Step Switch click
Sub StepChange

  CurStepIndex=CurStepIndex+1
  'Roll round CurStepIndex if greater than 3
  CurStepIndex=CurStepIndex Mod 6
  'Load the CurStepTW with the Step in TW
  LoadBCD CurStepTW(0),StepTWord$(CurStepIndex)

  'Could display something on the LCD to indicate New Step in force
  'eg Underline digit.
  DispFreq

End Sub
